@@ -1,10 +1,12 @@
+import sys
+import xml.etree.ElementTree as ET
+
 from collections import namedtuple
 from pathlib import Path
 from typing import Union
 from urllib.parse import urlparse
-import requests
-import xml.etree.ElementTree as ET
 
+import requests
 
 __copyright__ = 'Copyright 2021, 3Liz'
 __license__ = 'GPL version 3'
@@ -20,8 +22,11 @@ Plugin = namedtuple('Plugin', ['name', 'experimental', 'version'])
 
 class Merger:
 
+    """ To manage a merge between different XML files. """
+
     @classmethod
-    def xml_template(cls):
+    def xml_template(cls) -> str:
+        """ To create an empty XML files. """
         return (
             "<?xml version = '1.0' encoding = 'UTF-8'?>\n"
             "<plugins>\n"
@@ -29,31 +34,41 @@ class Merger:
         )
 
     def __init__(self, destination: Union[Path, None], input_uri: Union[Path, str, None]):
+        """ Constructor. """
+        # Usually a path, it can be None if we want to read only
         self.destination = destination
+        # Usually a path as a string, or a URL
         self.input_uri = input_uri
+
+        # Internal objects
         self.input_parser = None
         self.output_tree = None
         self.output_parser = None
 
     def exists(self) -> bool:
+        """ If the destination files exists. """
         return self.destination.exists()
 
     def init(self) -> None:
+        """ Init the XML files with an empty catalog. """
         print(f"Creating source {self.destination.absolute()}")
         with open(self.destination, 'w', encoding='utf8') as f:
             f.write(self.xml_template())
 
     def xml_input_parser(self) -> ET.Element:
+        """ Returns the XML parser for the input file. """
         if self.input_is_url():
             self.input_parser = ET.fromstring(requests.get(self.input_uri).content)
-            return self.input_parser
-
         else:
+            if not isinstance(self.input_uri, Path):
+                self.input_uri = Path(self.input_uri)
             tree = ET.parse(self.input_uri.absolute())
             self.input_parser = tree.getroot()
-            return self.input_parser
+
+        return self.input_parser
 
     def xml_output_parser(self) -> ET.Element:
+        """ Returns the XML parser for the output file. """
         if isinstance(self.destination, str):
             self.destination = Path(self.destination)
 
@@ -65,9 +80,14 @@ class Merger:
         return self.output_parser
 
     def input_is_url(self) -> bool:
+        """ Check if the input is a URL. """
         if isinstance(self.input_uri, Path):
             return False
 
+        path = Path(self.input_uri)
+        if path.exists():
+            return False
+
         # noinspection PyBroadException
         try:
             urlparse(self.input_uri)
@@ -95,8 +115,9 @@ class Merger:
         return plugins_list
 
     @staticmethod
-    def plugin_element(parser, name: str, experimental: bool) -> ET.Element:
-        for plugin in parser:
+    def plugin_element(parser: ET.Element, name: str, experimental: bool) -> [ET.Element, int]:
+        """ Return the XML element for a given plugin name and its experimental flag. """
+        for i, plugin in enumerate(parser):
             if plugin.attrib['name'] != name:
                 continue
 
@@ -107,10 +128,13 @@ class Merger:
                     if flag != experimental:
                         continue
 
-                    return plugin
+                    return plugin, i
+
+        return None, None
 
     @staticmethod
-    def diff_plugins(plugins_input, plugins_output):
+    def diff_plugins(plugins_input, plugins_output) -> list:
+        """ List of plugins which are in the input and not in the output. """
         new = []
         for plugin in plugins_input:
             if plugin not in plugins_output:
@@ -118,7 +142,12 @@ class Merger:
 
         return new
 
+    def count(self) -> int:
+        """ Count the number of plugins in the XML. """
+        return len(self.plugins(self.input_parser))
+
     def merge(self):
+        """ Make the merge. """
         if not self.exists():
             raise FileDoesNotExist
 
@@ -134,14 +163,20 @@ class Merger:
             print("No update is necessary")
 
         for plugin in diff:
-            # Remove old version
-            element = self.plugin_element(self.output_parser, plugin.name, plugin.experimental)
+            element, index = self.plugin_element(self.output_parser, plugin.name, plugin.experimental)
             if element:
-                print(f"Removing previous {plugin.name} {plugin.experimental} {element.attrib['version']}")
-                self.output_parser.remove(element)
+                print(f"Updating previous {plugin.name} {plugin.experimental} {element.attrib['version']}")
+                element.__setstate__(
+                    self.plugin_element(
+                        self.input_parser, plugin.name, plugin.experimental)[0].__getstate__())
+            else:
+                print(f"Adding new version {plugin.name} {plugin.experimental} {plugin.version}")
+                self.output_parser.append(
+                    self.plugin_element(self.input_parser, plugin.name, plugin.experimental)[0])
 
-            print(f"Adding new version {plugin.name} {plugin.experimental} {plugin.version}")
-            self.output_parser.append(
-                self.plugin_element(self.input_parser, plugin.name, plugin.experimental))
+        if sys.version_info >= (3, 9):
+            ET.indent(self.output_tree, space="\t", level=0)
+        self.output_tree.write(self.destination.absolute(), encoding="utf-8")
 
-        self.output_tree.write(self.destination.absolute())
+        with open(self.destination, "a", encoding='utf8') as f:
+            f.write("\n")
